/* $NetBSD$ */
/*-
 * Copyright (c) 2009 Elad Efrat <elad@NetBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/queue.h>
#include <sys/kmem.h>
#include <sys/errno.h>

#include "acl_core.h"

/* Return an initialized ACE spec. */
struct ace_spec *
ace_spec_new(void)
{
	struct ace_spec *spec;

	spec = kmem_zalloc(sizeof(*spec), KM_SLEEP);

	return spec;
}

void
ace_spec_free(struct ace_spec *spec)
{
	kmem_free(spec, sizeof(*spec));
}

void
ace_spec_setflags(struct ace_spec *spec, uint64_t flags)
{
	spec->spec_flags = flags;
}

struct ace_id *
ace_id_new(void)
{
	struct ace_id *id;

	id = kmem_zalloc(sizeof(*id), KM_SLEEP);

	return id;
}

void
ace_id_free(struct ace_id *id)
{
	kmem_free(id, sizeof(*id));
}

void
ace_id_set_uid(struct ace_id *id, uid_t uid)
{
	id->id_type = ACE_USER;
	id->id_uid = uid;
}

void
ace_id_set_gid(struct ace_id *id, gid_t gid)
{
	id->id_type = ACE_GROUP;
	id->id_gid = gid;
}

struct ace *
ace_new(void)
{
	struct ace *ace;

	ace = kmem_zalloc(sizeof(*ace), KM_SLEEP);

	return ace;
}

void
ace_free(struct ace *ace)
{
	/* Free specifications. */
	while (LIST_FIRST(&ace->ace_specs) != NULL) {
		struct ace_spec *spec;

		spec = LIST_FIRST(&ace->ace_specs);
		LIST_REMOVE(spec, spec_list);

		ace_spec_free(spec);
	}

	/* Free ACE identity. */
	ace_id_free(ace->ace_id);

	kmem_free(ace, sizeof(*ace));
}

void
ace_set_id(struct ace *ace, struct ace_id *id)
{
	ace->ace_id = id;
}

void
ace_add(struct ace *ace, struct ace_spec *spec)
{
	LIST_INSERT_HEAD(&ace->ace_specs, spec, spec_list);
}

struct acl *
acl_new(void)
{
	struct acl *acl;

	acl = kmem_zalloc(sizeof(*acl), KM_SLEEP);
	LIST_INIT(&acl->acl_aces);

	return acl;
}

void
acl_free(struct acl *acl)
{
	/* Free ACEs. */
	while (LIST_FIRST(&acl->acl_aces) != NULL) {
		struct ace *ace;

		ace = LIST_FIRST(&acl->acl_aces);
		LIST_REMOVE(ace, ace_list);

		ace_free(ace);
	}

	kmem_free(acl, sizeof(*acl));
}

int
acl_add(struct acl *acl, struct ace *ace)
{

	LIST_INSERT_HEAD(&acl->acl_aces, ace, ace_list);
	return 0;
}

/* Retrieve the ACE associated with a uid. */
struct ace *
acl_lookup_uid(struct acl *acl, uid_t uid)
{
	struct ace *ace = NULL;

	LIST_FOREACH(ace, &acl->acl_aces, ace_list) {
		if ((ace->ace_id->id_type == ACE_USER) &&
		    (ace->ace_id->id_uid == uid))
			return ace;
	}

	return NULL;
}

/* Retrieve the ACE associated with a gid. */
struct ace *
acl_lookup_gid(struct acl *acl, gid_t gid)
{
	struct ace *ace = NULL;

	LIST_FOREACH(ace, &acl->acl_aces, ace_list) {
		if ((ace->ace_id->id_type == ACE_GROUP) &&
		    (ace->ace_id->id_gid == gid))
			return ace;
	}

	return NULL;
}

int
acl_test_uid(struct acl *acl, uid_t uid, uint64_t flags)
{
	struct ace *ace;
	struct ace_spec *spec;

	ace = acl_lookup_uid(acl, uid);
	if (ace == NULL)
		return EACCES;

	LIST_FOREACH(spec, &ace->ace_specs, spec_list) {
		if (spec->spec_flags & flags) {
			return 0;
		}
	}

	return EACCES;
}

int
acl_test_gid(struct acl *acl, gid_t gid, uint64_t flags)
{
	struct ace *ace;
	struct ace_spec *spec;

	ace = acl_lookup_gid(acl, gid);
	if (ace == NULL)
		return EACCES;

	LIST_FOREACH(spec, &ace->ace_specs, spec_list) {
		if (spec->spec_flags & flags) {
			return 0;
		}
	}

	return EACCES;
}

