/* $NetBSD$ */
/*-
 * Copyright (c) 2009 Elad Efrat <elad@NetBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__KERNEL_RCSID(0, "$NetBSD$");

#include <sys/types.h>
#include <sys/param.h>
#include <sys/kauth.h>

#include <sys/sysctl.h>
#include <sys/vnode.h>
#include <sys/namei.h>
#include <sys/fileassoc.h>

#include <secmodel/acl/acl.h>
#include <secmodel/acl/acl_core.h>

#include <prop/proplib.h>

static fileassoc_t acl_fileassoc;
static kauth_listener_t l_vnode;

static int
spec_from_dict(prop_dictionary_t spec_dict, struct ace_spec **specp)
{
	struct ace_spec *spec;
	uint64_t flags = 0;
	bool b;

	spec = ace_spec_new();

	b = false;
	prop_dictionary_get_bool(spec_dict, "read", &b);
	if (b == true)
		flags |= ACE_READ;
	b = false;
	prop_dictionary_get_bool(spec_dict, "write", &b);
	if (b == true)
		flags |= ACE_WRITE;
	b = false;
	prop_dictionary_get_bool(spec_dict, "exec", &b);
	if (b == true)
		flags |= ACE_EXEC;

	ace_spec_setflags(spec, flags);

	*specp = spec;

	return 0;
}

static int
ace_from_dict(prop_dictionary_t ace_dict, struct ace **acep)
{
	prop_array_t specs;
	struct ace *ace;
	struct ace_id *id;
	const char *s;
	unsigned int i;

	s = NULL;
	prop_dictionary_get_cstring_nocopy(ace_dict, "identity_type", &s);
	if (s == NULL)
		return EINVAL;

	ace = ace_new();
	id = ace_id_new();

	/* Resolve ACE identity. */
	if (strcmp(s, "uid") == 0) {
		uid_t uid;

		uid = 0;
		prop_dictionary_get_uint32(ace_dict, "uid", &uid);
		ace_id_set_uid(id, uid);
	} else if (strcmp(s, "gid") == 0) {
		gid_t gid;

		gid = 0;
		prop_dictionary_get_uint32(ace_dict, "gid", &gid);
		ace_id_set_gid(id, gid);
	} else {
		ace_id_free(id);
		ace_free(ace);

		return EINVAL;
	}
	ace_set_id(ace, id);

	/* Set ACE specifications. */
	specs = prop_dictionary_get(ace_dict, "specs");
	for (i = 0; i < prop_array_count(specs); i++) {
		prop_dictionary_t spec_dict;
		struct ace_spec *spec = NULL;

		spec_dict = prop_array_get(specs, i);
		spec_from_dict(spec_dict, &spec);

		/* Add to ACE. */
		ace_add(ace, spec);
	}

	*acep = ace;

	return 0;
}

static int
acl_from_dict(prop_dictionary_t acl_dict, struct acl **aclp)
{
	prop_array_t aces;
	struct acl *acl;
	unsigned int i;

	/* Get the ACEs. */
	aces = prop_dictionary_get(acl_dict, "aces");
	if ((aces == NULL) || (prop_array_count(aces) == 0))
		return EINVAL;

	/* Initialize an ACL for the file. */
	acl = acl_new();

	for (i = 0; i < prop_array_count(aces); i++) {
		prop_dictionary_t ace_dict;
		struct ace *ace = NULL;

		ace_dict = prop_array_get(aces, i);
		ace_from_dict(ace_dict, &ace);

		/* Add ACE to ACL. */
		acl_add(acl, ace);
	}

	*aclp = acl;

	return 0;
}

int
secmodel_acl_set_sysctl(SYSCTLFN_ARGS)
{
	prop_dictionary_t dict;
	struct acl *acl = NULL;
	const char *s;
	struct nameidata nid;
	int error;
	struct vnode *vp;

	if (newp == NULL)
		return 0;

	dict = prop_dictionary_internalize(newp);
	if (dict == NULL)
		return EINVAL;

	/* Get the filename. */
	s = NULL;
	prop_dictionary_get_cstring_nocopy(dict, "filename", &s);
	if (s == NULL) {
		prop_object_release(dict);

		return EINVAL;
	}

	NDINIT(&nid, LOOKUP, FOLLOW, UIO_SYSSPACE, s);
	error = namei(&nid);
	if (error) {
		prop_object_release(dict);

		return error;
	}

	vp = nid.ni_vp;

	acl_from_dict(dict, &acl);

	fileassoc_add(vp, acl_fileassoc, acl);

	vrele(vp);

	return 0;
}

static int
spec_to_dict(struct ace_spec *spec, prop_dictionary_t *dictp)
{
	prop_dictionary_t dict;

	dict = prop_dictionary_create();

	if (spec->spec_flags & ACE_READ)
		prop_dictionary_set_bool(dict, "read", true);
	if (spec->spec_flags & ACE_WRITE)
		prop_dictionary_set_bool(dict, "write", true);
	if (spec->spec_flags & ACE_EXEC)
		prop_dictionary_set_bool(dict, "exec", true);

	*dictp = dict;

	return 0;
}

static int
ace_to_dict(struct ace *ace, prop_dictionary_t *dictp)
{
	prop_dictionary_t dict;
	prop_array_t specs;
	struct ace_spec *spec;

	dict = prop_dictionary_create();

	specs = prop_array_create();
	prop_dictionary_set(dict, "specs", specs);

	/* Add ACE specifications. */
	LIST_FOREACH(spec, &ace->ace_specs, spec_list) {
		prop_dictionary_t spec_dict;

		spec_to_dict(spec, &spec_dict);
		prop_array_add(specs, spec_dict);
	}

	/* Set identity. */
	if (ace->ace_id->id_type == ACE_USER) {
		prop_dictionary_set_uint32(dict, "uid", ace->ace_id->id_uid);
		prop_dictionary_set_cstring_nocopy(dict, "identity_type", "uid");
	} else if (ace->ace_id->id_type == ACE_GROUP) {
		prop_dictionary_set_uint32(dict, "gid", ace->ace_id->id_gid);
		prop_dictionary_set_cstring_nocopy(dict, "identity_type", "gid");
	} else {
		return EINVAL;
	}

	*dictp = dict;

	return 0;
}

static int
acl_to_dict(struct acl *acl, prop_dictionary_t *dictp)
{
	prop_dictionary_t dict;
	prop_array_t aces;
	struct ace *ace;

	dict = prop_dictionary_create();

	/* Create ACEs array. */
	aces = prop_array_create();
	prop_dictionary_set(dict, "aces", aces);

	LIST_FOREACH(ace, &acl->acl_aces, ace_list) {
		prop_dictionary_t ace_dict = NULL;

		ace_to_dict(ace, &ace_dict);
		prop_array_add(aces, ace_dict);
	}

	*dictp = dict;

	return 0;
}

int
secmodel_acl_get_sysctl(SYSCTLFN_ARGS)
{
	prop_dictionary_t dict;
	char *reqplist;
	struct acl *acl = NULL;
	const char *s;
	struct nameidata nid;
	int error;
	struct vnode *vp;
	char *aclplist;

	reqplist = kmem_zalloc(*oldlenp, KM_SLEEP);
	copyin(oldp, reqplist, *oldlenp);

	dict = prop_dictionary_internalize(reqplist);
	if (dict == NULL)
		return EINVAL;

	/* Get the filename. */
	s = NULL;
	prop_dictionary_get_cstring_nocopy(dict, "filename", &s);
	if (s == NULL) {
		prop_object_release(dict);

		return EINVAL;
	}

	NDINIT(&nid, LOOKUP, FOLLOW, UIO_SYSSPACE, s);
	error = namei(&nid);
	if (error) {
		prop_object_release(dict);

		return error;
	}

	vp = nid.ni_vp;

	/* Get ACL for file. */
	acl = fileassoc_lookup(vp, acl_fileassoc);
	if (acl == NULL) {
		prop_object_release(dict);
		vrele(vp);

		return ENOENT;
	}

	acl_to_dict(acl, &dict);

	/* Set filename on the result dictionary. */
	prop_dictionary_set_cstring(dict, "filename", s);

	aclplist = prop_dictionary_externalize(dict);

	if (newlen < strlen(aclplist) + 1) {
		return ENOMEM;
	}

	copyout(aclplist, __UNCONST(newp), strlen(aclplist) + 1); // XXX

	return 0;
}

SYSCTL_SETUP(sysctl_security_acl_setup,
    "sysctl security acl setup")
{
	const struct sysctlnode *rnode;

	sysctl_createv(clog, 0, NULL, &rnode,
		       CTLFLAG_PERMANENT,
		       CTLTYPE_NODE, "security", NULL,
		       NULL, 0, NULL, 0,
		       CTL_SECURITY, CTL_EOL);

	sysctl_createv(clog, 0, &rnode, &rnode,
		       CTLFLAG_PERMANENT,
		       CTLTYPE_NODE, "acl", NULL,
		       NULL, 0, NULL, 0,
		       CTL_CREATE, CTL_EOL);

	sysctl_createv(clog, 0, &rnode, NULL,
		       CTLFLAG_PERMANENT,
		       CTLTYPE_NODE, "set",
		       SYSCTL_DESCR("Set ACL"),
		       secmodel_acl_set_sysctl, 0, NULL, 0,
		       CTL_CREATE, CTL_EOL);

	sysctl_createv(clog, 0, &rnode, NULL,
		       CTLFLAG_PERMANENT,
		       CTLTYPE_NODE, "get",
		       SYSCTL_DESCR("Get ACL"),
		       secmodel_acl_get_sysctl, 0, NULL, 0,
		       CTL_CREATE, CTL_EOL);
}

void
secmodel_acl_start(void)
{
	fileassoc_register("acl", (fileassoc_cleanup_cb_t)acl_free,
	    &acl_fileassoc);

	l_vnode = kauth_listen_scope(KAUTH_SCOPE_VNODE,
	    secmodel_acl_vnode_cb, NULL);
}

#if defined(_LKM)
void
secmodel_acl_stop(void)
{
	kauth_unlisten_scope(l_vnode);

	fileassoc_deregister(acl_fileassoc);
}
#endif /* _LKM */

int
secmodel_acl_vnode_cb(kauth_cred_t cred,
    kauth_action_t action, void *cookie, void *arg0, void *arg1,
    void *arg2, void *arg3)
{
	int result;
	enum kauth_system_req req;
	struct vnode *vp, *dvp;
	struct acl *acl;
	uint64_t flags = 0;

	result = KAUTH_RESULT_DEFER;
	req = (enum kauth_system_req)arg0;

	vp = arg0;
	dvp = arg1;

	acl = fileassoc_lookup(vp, acl_fileassoc);
	if (acl == NULL)
		return result;

	if (action & KAUTH_VNODE_READ)
		flags |= ACE_READ;
	if (action & KAUTH_VNODE_WRITE)
		flags |= ACE_WRITE;
	if (action & KAUTH_VNODE_EXECUTE)
		flags |= ACE_EXEC;

	if (acl_test_uid(acl, kauth_cred_geteuid(cred), flags) == 0)
		result = KAUTH_RESULT_ALLOW;

	return result;
}

